/* - Ftp Server
 * Copyright (c) 1985, 1988, 1990, 1992, 1993, 1994, 2002
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if 0
static char sccsid[] = "@(#)ftpd.c	8.5 (Berkeley) 4/28/95";
#endif

/*
 * FTP server.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <alloca.h>

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_WAIT_H
#  include <sys/wait.h>
#endif

#include <netinet/in.h>
#ifdef HAVE_NETINET_IN_SYSTM_H
#  include <netinet/in_systm.h>
#endif
#ifdef HAVE_NETINET_IP_H
#  include <netinet/ip.h>
#endif

#define	FTP_NAMES
#include <arpa/ftp.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <limits.h>
#include <netdb.h>
#include <setjmp.h>
#include <signal.h>
#include <grp.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <fnmatch.h>

#include <tf_io.h>
#include <tf_util.h>
#include <daemon.h>

#include "extern.h"
#include "localhost.h"

#ifndef LINE_MAX
# define LINE_MAX 2048
#endif

#ifndef LOG_FTP
# define LOG_FTP LOG_DAEMON	/* Use generic facility.  */
#endif

#ifndef MAP_FAILED
# define MAP_FAILED (void*)-1
#endif

/* Exported to ftpcmd.h.  */
struct	sockaddr_in data_dest; /* Data port.  */
struct	sockaddr_in his_addr;  /* Peer address.  */
int	logging;               /* Enable log to syslog.  */
int	type = TYPE_A;         /* Default TYPE_A.  */
int	form = FORM_N;                  /* Default FORM_N.  */
int	debug;                 /* Enable debug mode if 1.  */
int	timeout = 300;         /* Timeout after 5 minutes of inactivity.  */
int	maxtimeout = 7200;     /* Don't allow idle time to be set
					beyond 2 hours.  */
int	pdata = -1;            /* For passive mode.  */
char	*hostname;             /* Who we are.  */
int	usedefault = 1;	       /* For data transfers.  */
char	tmpline[7];            /* Temp buffer use in OOB.  */
int	turbo = 0;

static int use_eplf = 1;

jmp_buf  errcatch;

/* Requester credentials.  */
struct credentials cred;

static struct  sockaddr_in ctrl_addr;    /* Control address.  */
static struct  sockaddr_in data_source;  /* Port address.  */
static struct  sockaddr_in pasv_addr;    /* Pasv address.  */

static int      data = -1;       /* Port data connection socket.  */
static jmp_buf  urgcatch;
static int      stru = STRU_F;     /* Avoid C keyword.  */
static int      stru_mode = MODE_S;/* Default STRU mode stru_mode = MODE_S.  */
static int      anon_only;       /* Allow only anonymous login.  */
static int      no_version;      /* Don't print version to client.  */
static int      daemon_mode;     /* Start in daemon mode.  */
static off_t    file_size;
static off_t    byte_count;
static sig_atomic_t transflag;   /* Flag where in a middle of transfer.  */
static const char *pid_file = PATH_FTPDPID;
#if !defined(CMASK) || CMASK == 0
#undef CMASK
#define CMASK 027
#endif
static int  defumask = CMASK;    /* Default umask value.  */
static int login_attempts;       /* Number of failed login attempts.  */
static int askpasswd;		 /* Had user command, ask for passwd.  */
static char curname[10];	 /* Current USER name.  */
static tf_handle tf;

/*
 * Timeout intervals for retrying connections
 * to hosts that don't accept PORT cmds.  This
 * is a kludge, but given the problems with TCP...
 */
#define	SWAITMAX	90	/* wait at most 90 seconds */
#define	SWAITINT	5	/* interval between retries */

static int swaitmax = SWAITMAX;
static int swaitint = SWAITINT;

#ifdef HAVE_SETPROCTITLE
char	proctitle[LINE_MAX];	/* initial part of title */
#endif /* SETPROCTITLE */

static void ack __P ((const char *));
#ifdef HAVE_LIBWRAP
static int  check_host __P ((struct sockaddr *sa));
#endif
static void complete_login __P ((struct credentials *));
static const char *curdir __P ((void));
static FILE *dataconn __P ((const char *, off_t, const char *));
static void dolog __P ((struct sockaddr_in *, struct credentials *));
static void end_login __P ((struct credentials *));
static FILE *getdatasock __P ((const char *));
static void lostconn __P ((int));
static void myoob __P ((int));
static void sigquit __P ((int));
static void usage __P ((const char *progname, int err));

static const char *short_options = "Aa:bDdehlp:qt:T:u:P:v";
static struct option long_options[] =
{
	{ "anonymous-only", no_argument, 0, 'A' },
	{ "auth", required_argument, 0, 'a' },
	{ "turbo", no_argument, 0, 'b' },
	{ "daemon", no_argument, 0, 'D' },
	{ "debug", no_argument, 0, 'd' },
	{ "noeplf", no_argument, 0, 'e' },
	{ "help", no_argument, 0, 'h' },
	{ "logging", no_argument, 0, 'l' },
	{ "pidfile", required_argument, 0, 'p' },
	{ "port", required_argument, 0, 'P' },
	{ "no-version", no_argument, 0, 'q' },
	{ "timeout", required_argument, 0, 't' },
	{ "max-timeout", required_argument, 0, 'T' },
	{ "umask", required_argument, 0, 'u' },
	{ "verbose", no_argument, 0, 'v' },
	{ "version", no_argument, 0, 'V' },
	{ 0, 0, 0, 0 }
};

static void LOGCMD(const char *cmd, const char *file)
{
	if (logging > 1) {
		syslog(LOG_INFO,"%s %s%s", cmd, *(file) == '/' ? "" : curdir(), file);
	}
}

static void LOGCMD2(const char *cmd, const char *file1, const char *file2)
{
	if (logging > 1) {
		syslog(LOG_INFO,"%s %s%s %s%s", cmd, *(file1) == '/' ? "" : curdir(), file1, *(file2) == '/' ? "" : curdir(), file2);
	}
}

static void LOGBYTES(const char *cmd, const char *file, off_t cnt)
{
	if (logging > 1) { 
			if (cnt == (off_t)-1)  {
				syslog(LOG_INFO,"%s %s%s", cmd, *(file) == '/' ? "" : curdir(), file); 
			}
			else  {
				syslog(LOG_INFO, "%s %s%s = %Ld bytes", cmd, (*(file) == '/') ? "" : curdir(), file, cnt); 
			}
	}
}

static void
usage (const char *progname, int err)
{
	if (err != 0)
		{
			fprintf (stderr, "Usage: %s [OPTION] ...\n", progname);
			fprintf (stderr, "Try `%s --help' for more information.\n", progname);
		}
	else
		{
			fprintf (stdout, "Usage: %s [OPTION] ...\n", progname);
			puts ("Internet File Transfer Protocol server.\n\n\
    -A, --anonymous-only      Server configure for anonymous service only\n\
    -b, --turbo               Default to turbo mode on for transfers\n\
    -D, --daemon              Start the ftpd standalone\n\
    -d, --debug               Debug mode\n\
    -e, --noeplf              Don't send listing in EPLF mode\n\
    -l, --logging             Increase verbosity of syslog messages\n\
    -p, --pidfile=[PIDFILE]   Change default location of pidfile\n\
    -P, --port=[port]         Change the port to listen on in server mode\n\
    -q, --no-version          Do not display version in banner\n\
    -t, --timeout=[TIMEOUT]   Set default idle timeout\n\
    -T, --max-timeout         Reset maximum value of timeout allowed\n\
    -u, --umask               Set default umask(base 8)\n\
            --help                Print this message\n\
    -V, --version             Print version\n\
    -a, --auth=[AUTH]         Use AUTH for authentication, it can be:\n\
                                 default     passwd authentication.");
		}
	exit (err);
}

int
main(int argc, char *argv[], char **envp)
{
	extern void yyparse __P((void));
	int option;
	int port = -1;
	const char *progname = argv[0];
	int verbose = 0;

#ifdef HAVE_TZSET
	tzset(); /* In case no timezone database in ~ftp.  */
#endif

#ifdef HAVE_INITSETPROCTITLE
	/* Save start and extent of argv for setproctitle.  */
	initsetproctitle (argc, argv, envp);
#endif /* HAVE_INITSETPROCTITLE */

	while ((option = getopt_long (argc, argv, short_options,
				long_options, NULL)) != EOF)
		{
			switch (option)
	{
	case 'A': /* Anonymous ftp only.  */
		anon_only = 1;
		break;

	case 'a': /* Authentification method.  */
		if (strcasecmp (optarg, "default") == 0)
			cred.auth_type = AUTH_TYPE_PASSWD;
#ifdef WITH_PAM
		else if (strcasecmp (optarg, "pam") == 0)
			cred.auth_type = AUTH_TYPE_PAM;
#endif
#ifdef WITH_KERBEROS
		else if (stracasecmp (optarg, "kerberos") == 0)
			cred.auth_type = AUTH_TYPE_KERBEROS;
#endif
#ifdef WITH_KERBEROS5
		else if (stracasecmp (optarg, "kerberos5") == 0)
			cred.auth_type = AUTH_TYPE_KERBEROS5;
#endif
#ifdef WITH_OPIE
		else if (stracasecmp (optarg, "opie") == 0)
			cred.auth_type = AUTH_TYPE_OPIE;
#endif
		break;

	case 'D': /* Run ftpd as daemon.  */
		daemon_mode = 1;
		break;

	case 'd': /* Enable debug mode.  */
		debug = 1;
		break;

	case 'e': /* Disable eplf listings */
		use_eplf = 0;
		break;

	case 'l': /* Increase logging level.  */
		logging++;	/* > 1 == Extra logging.  */
		break;

	case 'b': /* High speed (turbo) mode.  */
		turbo = 1;
		break;

	case 'p': /* Override pid file */
		pid_file = optarg;
		break;

	case 'P': /* Override default port */
		port = atoi(optarg);
		break;

	case 'q': /* Don't include version number in banner.  */
		no_version = 1;
		break;

	case 't': /* Set default timeout value.  */
		timeout = atoi (optarg);
		if (maxtimeout < timeout)
			maxtimeout = timeout;
		break;

	case 'T': /* Maximum timeout allowed.  */
		maxtimeout = atoi (optarg);
		if (timeout > maxtimeout)
			timeout = maxtimeout;
		break;

	case 'u': /* Set umask.  */
		{
			long val = 0;

			val = strtol (optarg, &optarg, 8);
			if (*optarg != '\0' || val < 0)
				fprintf (stderr, "%s: bad value for -u", argv[0]);
			else
				defumask = val;
			break;
		}

	case 'h': /* Usage.  */
		usage (progname, 0);
		/* Not reached.  */

	case 'V': /* Version.  */
		printf ("ftpd (%s) %s\n", PACKAGE_NAME, PACKAGE_VERSION);
		exit (0);

	case 'v': /* Increment verbosity level */
		verbose++;
		break;

	case '?':
	default:
		usage (progname, 1);
		/* Not reached.  */
	}
		}

	/* Bail out, wrong usage */
	argc -= optind;
	if (argc != 0)
		usage (progname, 1);

	/* LOG_NDELAY sets up the logging connection immediately,
		 necessary for anonymous ftp's that chroot and can't do it later.  */
	openlog ("ftpd", LOG_PID | LOG_NDELAY, LOG_FTP);

	if (!debug) {
		(void) freopen (PATH_DEVNULL, "w", stderr);
	}

	/* If not running via inetd, we detach and dup(fd, 0), dup(fd, 1) the
		 fd = accept(). tcpd is check if compile with the support  */
	if (daemon_mode)
		{
			if (server_mode (pid_file, &his_addr, port, debug) < 0)
	exit (1);
		}
	else
		{
			int addrlen = sizeof (his_addr);
			if (getpeername (STDIN_FILENO, (struct sockaddr *)&his_addr,
					 &addrlen) < 0)
	{
		syslog (LOG_ERR, "getpeername (%s): %m", progname);
		exit (1);
	}
		}

	(void) signal (SIGHUP, sigquit);
	(void) signal (SIGINT, sigquit);
	(void) signal (SIGQUIT, sigquit);
	(void) signal (SIGTERM, sigquit);
	(void) signal (SIGPIPE, lostconn);
	(void) signal (SIGCHLD, SIG_IGN);
	if (signal (SIGURG, myoob) == SIG_ERR)
		syslog (LOG_ERR, "signal: %m");

	/* Get info on the ctrl connection.  */
	{
		int addrlen = sizeof (ctrl_addr);
		if (getsockname (STDIN_FILENO, (struct sockaddr *)&ctrl_addr, &addrlen) < 0) {
			syslog (LOG_ERR, "getsockname (%s): %m", progname);
			exit (1);
		}
	}

#if defined (IP_TOS) && defined (IPTOS_LOWDELAY) && defined (IPPROTO_IP)
	/* To  minimize delays for interactive traffic.  */
	{
		int tos = IPTOS_LOWDELAY;
		if (setsockopt (STDIN_FILENO, IPPROTO_IP, IP_TOS,
				(char *)&tos, sizeof(int)) < 0)
			syslog (LOG_WARNING, "setsockopt (IP_TOS): %m");
	}
#endif

#ifdef SO_OOBINLINE
	/* Try to handle urgent data inline.  */
	{
		int on = 1;
		if (setsockopt (STDIN_FILENO, SOL_SOCKET, SO_OOBINLINE,
				(char *)&on, sizeof (on)) < 0)
			syslog (LOG_ERR, "setsockopt: %m");
	}
#endif

#ifdef SO_KEEPALIVE
	/* Set keepalives on the socket to detect dropped connections.  */
	{
		int keepalive = 1;
		if (setsockopt (STDIN_FILENO, SOL_SOCKET, SO_KEEPALIVE,
				(char *)&keepalive, sizeof (keepalive)) < 0)
			syslog (LOG_WARNING, "setsockopt (SO_KEEPALIVE): %m");
	}
#endif

#ifdef	F_SETOWN
	if (fcntl (STDIN_FILENO, F_SETOWN, getpid ()) == -1)
		syslog (LOG_ERR, "fcntl F_SETOWN: %m");
#endif

	dolog (&his_addr, &cred);

	/* Deal with login disable.  */
	if (display_file (PATH_NOLOGIN, 530) == 0)
		{
			reply (530, "System not available.");
			exit (0);
		}

	/* Display a Welcome message if exists,
		 N.B. reply(220,) must follow.  */
	(void) display_file (PATH_FTPWELCOME, 220);

	hostname = localhost ();
	if (! hostname)
		perror_reply (550, "Local resource failure: malloc");

	if (daemon_lock_pid("/tmp/toppy") != 0) {
		reply(550, "Topfield is already in use");
		exit(0);
	}

	/* REVISIT: trace level, timeout, devindex */
	memset(&tf, 0, sizeof(tf));
	tf.trace_level = verbose;

	if (topfield_open(&tf, 0) != 0) {
			reply(550, "Failed to connect to Topfield");
			exit(0);
	}

	/* Tell them we're ready to roll.  */
	if (!no_version)
		reply (220, "%s FTP server (%s %s) ready.",
		 hostname, PACKAGE_NAME, PACKAGE_VERSION);
	else
		reply (220, "%s FTP server ready.", hostname);

#if 0
	if (!tf.dev) {
		reply (220, "Running in fake mode");
	}
#endif

	/* Set the jump, if we have an error parsing,
		 come here and start fresh.  */
	(void) setjmp (errcatch);

	/* Roll.  */
	for (;;)
		(void) yyparse ();
	/* NOTREACHED */
}

static char *topfield_cwd;

void cwd(const char *newpath)
{
	tf_dirent dirent;

	char *newcwd = tf_makename(topfield_cwd ?: "", newpath);

	/* Now see if this is a valid directory */

	if (tf_stat(&tf, newcwd, &dirent) != 0 || dirent.type != 'd') {
			reply(550, "Failed to chdir to %s", newpath);
	}
	else {
		char *pt;

		ack ("CWD");

		free(topfield_cwd);
		topfield_cwd = strdup(newcwd);

		/* Now we translated \ to / */
		for (pt = topfield_cwd; *pt; pt++) {
			if (*pt == '\\') {
				*pt = '/';
			}
		}
	}
}

static const char *curdir(void)
{
	return topfield_cwd ?: "/";
}

time_t filetime(const char *path)
{
	tf_dirent dirent;

	char *fullpath = tf_makename(curdir(), path);


	if (tf_stat(&tf, fullpath, &dirent) != 0) {
		return -1;
	}
	else {
		return dirent.stamp;
	}
}

off_t filesize(const char *path)
{
	tf_dirent dirent;

	char *fullpath = tf_makename(curdir(), path);

	if (tf_stat(&tf, fullpath, &dirent) != 0 || dirent.type != 'f') {
		return -1;
	}
	else {
		return dirent.size;
	}
}

static void
sigquit (int signo)
{
#ifdef HAVE_STRSIGNAL
	syslog (LOG_ERR, "got signal %s", strsignal (signo));
#else
	syslog (LOG_ERR, "got signal %d", signo);
#endif
	dologout (-1);
}


static void
lostconn (int signo)
{
	(void)signo;
	if (debug)
		syslog (LOG_DEBUG, "lost connection");
	dologout (-1);
}

/* Helper function.  */
char *
sgetsave (const char *s)
{
	char *string;
	size_t len;

	if (s == NULL)
		s = "";

	len = strlen (s) + 1;
	string = malloc (len);
	if (string == NULL)
		{
			perror_reply (421, "Local resource failure: malloc");
			dologout (1);
			/* NOTREACHED */
		}
	/*  (void) strcpy (string, s); */
	memcpy (string, s, len);
	return string;
}

static void
complete_login (struct credentials *pcred)
{
	/* Display a login message, if it exists.
		N.B. reply(230,) must follow the message.  */
	(void) display_file (PATH_FTPLOGINMESG, 230);

	if (pcred->guest)
		{
			reply (230, "Guest login ok, access restrictions apply.");
#ifdef HAVE_SETPROCTITLE
			snprintf (proctitle, sizeof (proctitle), "%s: anonymous",
		pcred->remotehost);
			setproctitle ("%s",proctitle);
#endif /* HAVE_SETPROCTITLE */
			if (logging)
	syslog (LOG_INFO, "ANONYMOUS FTP LOGIN FROM %s",
		pcred->remotehost);
		}
	else
		{
			reply (230, "User %s logged in.", pcred->name);
#ifdef HAVE_SETPROCTITLE
			snprintf (proctitle, sizeof (proctitle),
		"%s: %s", pcred->remotehost, pcred->name);
			setproctitle ("%s",proctitle);
#endif /* HAVE_SETPROCTITLE */
			if (logging)
	syslog (LOG_INFO, "FTP LOGIN FROM %s as %s",
		pcred->remotehost, pcred->name);
		}
	(void) umask(defumask);
	return;
}

/* USER command.
	 Sets global passwd pointer pw if named account exists and is acceptable;
	 sets askpasswd if a PASS command is expected.  If logged in previously,
	 need to reset state.  */
void
user (const char *name)
{
	if (cred.logged_in)
		{
			if (cred.guest)
	{
		reply (530, "Can't change user from guest login.");
		return;
	}
			end_login (&cred);
		}

	/* Non zero means failed.  */
	if (auth_user (name, &cred) != 0)
		{
			/* If they gave us a reason.  */
			if (cred.message)
	{
		reply (530, "%s", cred.message);
		free (cred.message);
		cred.message = NULL;
	}
			else
	reply (530, "User %s access denied.", name);
			if (logging)
	syslog(LOG_NOTICE, "FTP LOGIN REFUSED FROM %s, %s",
				 cred.remotehost, name);
			return;
		}

	/* If the server is set to serve anonymous service only
		 the request have to come from a guest or a chrooted.  */
	if (anon_only && !cred.guest && !cred.dochroot)
		{
			reply (530, "Sorry, only anonymous ftp allowed");
			return;
		}

	if (logging)
		{
			strncpy (curname, name, sizeof (curname) - 1);
			curname [sizeof (curname) - 1] = '\0'; /* Make sure null terminated.  */
		}

	if (cred.message)
		{
			reply (331, "%s", cred.message);
			free (cred.message);
			cred.message = NULL;
		}
	else
		reply (331, "Password required for %s.", name);

	askpasswd = 1;

	/* Delay before reading passwd after first failed
		 attempt to slow down passwd-guessing programs.  */
	if (login_attempts)
		sleep ((unsigned) login_attempts);
}

/* Terminate login as previous user, if any, resetting state;
	 used when USER command is given or login fails.  */
static void
end_login (struct credentials *pcred)
{
	char *remotehost = pcred->remotehost;
	int atype = pcred->auth_type;

	if (pcred->name)
		free (pcred->name);
	if (pcred->passwd)
		{
			memset (pcred->passwd, 0, strlen (pcred->passwd));
			free (pcred->passwd);
		}
	if (pcred->homedir)
		free (pcred->homedir);
	if (pcred->rootdir)
		free (pcred->rootdir);
	if (pcred->shell)
		free (pcred->shell);
	if (pcred->pass) /* ??? */
		{
			memset (pcred->pass, 0, strlen (pcred->pass));
			free (pcred->pass);
		}
	if (pcred->message)
		free (pcred->message);
	memset (pcred, 0, sizeof (*pcred));
	pcred->remotehost = remotehost;
	pcred->auth_type = atype;
}

void
pass (const char *passwd)
{
	if (cred.logged_in || askpasswd == 0)
		{
			reply(503, "Login with USER first.");
			return;
		}
	askpasswd = 0;

	if (!cred.guest) /* "ftp" is the only account allowed no password.  */
		{
			/* Try to authenticate the user.  Failed if != 0.  */
			if (auth_pass (passwd, &cred) != 0)
	{
		/* Any particular reasons.  */
		if (cred.message)
			{
				reply (530, "%s", cred.message);
				free (cred.message);
				cred.message = NULL;
			}
		else
			reply (530, "Login incorrect.");
		if (logging)
			syslog (LOG_NOTICE, "FTP LOGIN FAILED FROM %s, %s",
				cred.remotehost, curname);
		if (login_attempts++ >= 5)
			{
				syslog(LOG_NOTICE, "repeated login failures from %s",
				 cred.remotehost);
				exit(0);
			}
		return;
	}
		}
	cred.logged_in = 1; /* Everything seems to be allright.  */
	complete_login (&cred);
	login_attempts = 0; /* This time successful.  */
}

void
retrieve (const char *cmd, const char *name)
{
	FILE *dout = 0;
	off_t current = restart_point;
	tf_dirent dirent;
	tf_buffer buf;
	int ret;
	int retries = 0;


	char *srcpath = tf_makename(curdir(), name);

	if (turbo) {
		tf_cmd_turbo(&tf, 1);
	}

resume_get:

	if (retries++ >= 5) {
		if (debug) {
			fprintf(stderr, "retrieve() -- too many retries, giving up\n");
		}
		perror_reply (550, name);
		if (turbo) {
			tf_cmd_turbo(&tf, 0);
		}
		return;
	}

	ret = tf_cmd_get(&tf, srcpath, current, &dirent);

	/*fprintf(stderr, "tf_cmd_get() srcpath=%s, current=%Ld, ret=%d\n", srcpath, current, ret);*/

	if (ret != 0) {
		if (debug) {
			fprintf(stderr, "tf_cmd_get() returned %d\n", ret);
		}
		goto resume_get;
	}

	if (!dout) {
		dout = dataconn (name, dirent.size, "w");

		if (dout == NULL) {
			fprintf(stderr, "Failed to open dout\n");
			tf_cmd_get_cancel(&tf);
		}
		else {
			/* Put dout into nonblocking mode */
			long on = 1;
			fcntl(fileno(dout), F_SETFL, O_NONBLOCK, &on);
		}
	}
	if (dout) {
		/*fprintf(stderr, "Reading data\n");*/
		while ((ret = tf_cmd_get_next(&tf, &buf)) == 0) {
			int w = 0;

			if (debug) {
				/*fprintf(stderr, "Read %d bytes @%lld, writing @%lld\n", buf.size, buf.offset, current);*/
			}
			if (buf.offset == current) {
				/* This is what we expected */
				current += buf.size;
			}
			else {
				if (debug) {
					fprintf(stderr, "Read %d bytes offset @%lld != writing @%lld\n", buf.size, buf.offset, current);
				}
				syslog(LOG_ERR, "Read %d bytes offset @%lld != writing @%lld", buf.size, buf.offset, current);
				if (buf.offset > current) {
					syslog(LOG_ERR, "Missed some data, so retrying");
					/* Just ask the topfield for the correct data again */
					goto resume_get;
				}
				if (buf.offset < current) {
					/* Duplicate data, so discard some or all of it */
					if (buf.offset + buf.size <= current) {
						syslog(LOG_ERR, "Discarding completely duplicate data");
						/* Just discard this data */
						continue;
					}

					/* Need to keep just some of it */
					syslog(LOG_ERR, "Discarding some duplicate data");

					w = current - buf.offset;
					current = buf.size + buf.offset;
				}
			}


			while (w < buf.size) {
				fd_set fds;
				struct timeval to;
				int n;

				/* Set up for the select */
				FD_ZERO(&fds);
				FD_SET(fileno(dout), &fds);

				/* Wait up to 2 seconds for dout to be writable.
				 * If we wait too long, the topfield gets stuck
				 * and won't talk to us any more :-(
				 */
				to.tv_sec = 2;
				to.tv_usec = 0;

				/* Wait for a response */
				n = select(fileno(dout) + 1, 0, &fds, 0, &to);
				if (n == 0) {
					if (debug) {
						fprintf(stderr, "Timeout waiting to write to client\n");
					}
					break;
				}
				if (n < 0) {
					if (debug) {
						fprintf(stderr, "select returned %d: %m\n", n);
					}
					break;
				}

				/*fprintf(stderr, "Output is now writable\n");*/
				n = write(fileno(dout), buf.data + w, buf.size - w);
				/*fprintf(stderr, "Wrote %d bytes to dout\n", n);*/
				if (n <= 0) {
					if (debug) {
						fprintf(stderr, "write returned %d: %m\n", n);
					}
					break;
				}

				w += n;
			}

			if (w != buf.size) {
				if (debug) {
					fprintf(stderr, "Wrote %d of %d bytes: %m\n", w, buf.size);
				}
				ret = -200;
				break;
			}
			else {
				/* All OK, so reduce the retries */
				if (retries) {
					retries--;
				}
			}
		}
		if (debug) {
			fprintf(stderr, "At end of loop, ret=%d\n", ret);
		}

		if (ret <= 0) {
			/* Cancel anything which may be outstanding */
			if (tf_cmd_get_cancel(&tf) != 0) {
				fprintf(stderr, "Failed to cancel after ret=%d\n", ret);
			}
			else if (ret == TF_ERR_IO) {
				if (debug) {
					fprintf(stderr, "Resuming interrupted get @%lld %s\n", current, srcpath);
				}
				goto resume_get;
			}
			if (debug) {
				if (ret == TF_ERR_IO) {
					fprintf(stderr, "Failed to resume interrupted get\n");
				}
				else {
					fprintf(stderr, "Cancelled get\n");
				}
			}
		}
	}

	if (turbo) {
		tf_cmd_turbo(&tf, 0);
	}

	/*fprintf(stderr, "Finishing up and closing data connection\n");*/
	if (ret >= 0) {
		if (debug) {
			fprintf(stderr, "Transfer succeeded %lld bytes %s\n", current, srcpath);
		}
		reply (226, "Transfer complete.");
	}
	else {
		if (debug) {
			fprintf(stderr, "Transfer failed %lld bytes %s\n", current, srcpath);
		}
		perror_reply (550, name);
	}

	if (dout) {
		fclose(dout);
	}
	data = -1;
	pdata = -1;
}

void
store (const char *name, const char *mode, int unique)
{
	FILE *din = 0;
	off_t current = restart_point;
	int ret;

	char *destpath = tf_makename(curdir(), name);

	if (unique) {
		reply(220, "Ignoring unique!");
	}

	if (turbo) {
		tf_cmd_turbo(&tf, 1);
	}

	if (current == 0) {
		/* Need to delete any existing file since it won't just be truncated */
		tf_cmd_delete(&tf, destpath);
	}

resume_put:

	ret = tf_cmd_put(&tf, destpath, 0, time(0), current);

#ifdef DEBUG
	fprintf(stderr, "tf_cmd_put(destpath=%s, restart_point=%lld) = %d\n", destpath, current, ret);
#endif

	if (ret != 0) {
		perror_reply (553, name);

		LOGCMD (*mode == 'w' ? "put" : "append", name);
		if (turbo) {
			tf_cmd_turbo(&tf, 0);
		}
		return;
	}

	if (!din) {
		din = dataconn (name, (off_t)-1, "r");
		if (din == NULL) {
			tf_cmd_put_cancel(&tf);
		}
	}

	if (din) {
		char buf[MAX_PUT_SIZE];

		for (;;) {
			fd_set fds;
			struct timeval to;

			/* Set up for the select */
			FD_ZERO(&fds);
			FD_SET(fileno(din), &fds);

			/* Wait up to 5 seconds for din to be readable.
			 */
			to.tv_sec = 5;
			to.tv_usec = 0;

			/* Wait for a response */
			/*fprintf(stderr, "put: waiting for data\n");*/
			ret = select(fileno(din) + 1, &fds, 0, 0, &to);
			if (ret > 0) {
				int r;

				/*fprintf(stderr, "Input is now readable\n");*/
				ret = read(fileno(din), buf, sizeof(buf));

				/*fprintf(stderr, "put: read %d bytes - sending\n", ret);*/

				if (ret > 0) {
					r = tf_cmd_put_data(&tf, current, buf, ret);

					/*fprintf(stderr, "put: tf_cmd_put_data() returned %d\n", r);*/
					if (r != 0) {
						ret = r;
						break;
					}
					current += ret;
					byte_count += ret;
				}
			}
			if (ret <= 0 && errno != EINTR) {
				break;
			}
		}
		
		if (ret >= 0) {
			/*fprintf(stderr, "put: tf_cmd_put_done()\n");*/
			ret = tf_cmd_put_done(&tf);
			reply (226, "Transfer complete.");
		}
		else {
			/*fprintf(stderr, "put: tf_cmd_put_cancel()\n");*/

			/* Cancel anything which may be outstanding */
			if (tf_cmd_put_cancel(&tf) == 0 && ret == TF_ERR_IO) {
				if (debug) {
					fprintf(stderr, "Resuming interrupted put @%lld %s\n", current, name);
				}
				goto resume_put;
			}

			if (debug) {
				if (ret == TF_ERR_IO) {
					fprintf(stderr, "Failed to resume interrupted put\n");
				}
				else {
					fprintf(stderr, "Cancelled put\n");
				}
			}

			/* Failed to cancel the put, so give up */
			perror_reply (426, "Data Connection");
		}
		fclose(din);
	}

	if (turbo) {
		tf_cmd_turbo(&tf, 0);
	}

	data = -1;
	pdata = -1;

	LOGBYTES (*mode == 'w' ? "put" : "append", name, byte_count);

}

static FILE *
getdatasock (const char *mode)
{
	int s, t, tries;

	if (data >= 0)
		return fdopen (data, mode);
	(void) seteuid ((uid_t)0);
	s = socket (AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		goto bad;

	/* Enables local reuse address.  */
	{
		int on = 1;
		if (setsockopt (s, SOL_SOCKET, SO_REUSEADDR,
				(char *) &on, sizeof(on)) < 0)
			goto bad;
	}

	/* anchor socket to avoid multi-homing problems */
	data_source.sin_family = AF_INET;
	data_source.sin_addr = ctrl_addr.sin_addr;
	for (tries = 1; ; tries++)
		{
			if (bind (s, (struct sockaddr *)&data_source,
		sizeof(data_source)) >= 0)
	break;
			if (errno != EADDRINUSE || tries > 10)
	goto bad;
			sleep (tries);
		}
	(void) seteuid ((uid_t)cred.uid);

#if defined (IP_TOS) && defined (IPTOS_THROUGHPUT) && defined (IPPROTO_IP)
	{
		int on = IPTOS_THROUGHPUT;
		if (setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int)) < 0)
			syslog(LOG_WARNING, "setsockopt (IP_TOS): %m");
	}
#endif

	return (fdopen(s, mode));
 bad:
	/* Return the real value of errno (close may change it) */
	t = errno;
	(void) seteuid ((uid_t)cred.uid);
	(void) close(s);
	errno = t;
	return NULL;
}

static FILE *
dataconn (const char *name, off_t size, const char *mode)
{
	char sizebuf[32];
	FILE *file;
	int retry = 0;

	file_size = size;
	byte_count = 0;
	if (size != (off_t) -1)
		(void) snprintf(sizebuf, sizeof(sizebuf), " (%Ld bytes)",
				size);
	else
		*sizebuf = '\0';
	if (pdata >= 0)
		{
			struct sockaddr_in from;
			int s, fromlen = sizeof (from);

			(void) signal (SIGALRM, toolong);
			(void) alarm ((unsigned) timeout);
			s = accept (pdata, (struct sockaddr *)&from, &fromlen);
			(void) alarm (0);
			if (s < 0)
	{
		reply(425, "Can't open data connection.");
		(void) close (pdata);
		pdata = -1;
		return NULL;
	}
			(void) close (pdata);
			pdata = s;
#if defined (IP_TOS) && defined (IPTOS_THROUGHPUT) && defined (IPPROTO_IP)
			/* Optimize throughput.  */
			{
	int tos = IPTOS_THROUGHPUT;
	(void) setsockopt (s, IPPROTO_IP, IP_TOS, (char *)&tos, sizeof (int));
			}
#endif
#ifdef SO_KEEPALIVE
			/* Set keepalives on the socket to detect dropped conns.  */
			{
	int keepalive = 1;
	(void) setsockopt (s, SOL_SOCKET, SO_KEEPALIVE,
				 (char *)&keepalive, sizeof (int));
			}
#endif
			reply (150, "Opening %s mode data connection for '%s'%s.",
			 type == TYPE_A ? "ASCII" : "BINARY", name, sizebuf);
			return fdopen (pdata, mode);
		}
	if (data >= 0)
		{
			reply (125, "Using existing data connection for '%s'%s.",
			 name, sizebuf);
			usedefault = 1;
			return fdopen (data, mode);
		}
	if (usedefault)
		data_dest = his_addr;
	usedefault = 1;
	file = getdatasock (mode);
	if (file == NULL)
		{
			reply (425, "Can't create data socket (%s,%d): %s.",
			 inet_ntoa (data_source.sin_addr),
			 ntohs (data_source.sin_port), strerror(errno));
			return NULL;
		}
	data = fileno (file);
	while (connect (data, (struct sockaddr *)&data_dest,
			sizeof (data_dest)) < 0)
		{
			if (errno == EADDRINUSE && retry < swaitmax)
	{
		sleep ((unsigned) swaitint);
		retry += swaitint;
		continue;
	}
			perror_reply (425, "Can't build data connection");
			(void) fclose (file);
			data = -1;
			return NULL;
		}
	reply (150, "Opening %s mode data connection for '%s'%s.",
	 type == TYPE_A ? "ASCII" : "BINARY", name, sizebuf);
	return file;
}

static void dump_dirent(FILE *fh, const char *filename, tf_dirent *dirent)
{
	const char *mode = (dirent->type == 'f') ? "-rw-r--r--" : "drwxr-xr-x";
	struct tm *tm = localtime(&dirent->stamp);
	char timebuf[24];
	strftime(timebuf, sizeof(timebuf), "%b %e %Y ", tm);

	fprintf(fh, "%s   1 none none %12llu %s %s", mode, dirent->size, timebuf, filename ?: dirent->name);
}

/*
 * From lftp source
 
+i774.71425,m951188401,/,	users
+i774.49602,m917883130,r,s79126,	jgr_www2.exe

starts with +
comma separated
first character of field is type:
 i - ?
 m - modification time
 / - means directory
 r - means plain file
 s - size
 up - permissions in octal
 \t - file name follows.
*/

/**
 * Dump in EPLF format.
 */
static void dump_dirent_as_eplf(FILE *fh, const char *filename, tf_dirent *dirent)
{
	fprintf(fh, "+");

	if (dirent->stamp) {
		fprintf(fh, "m%lu,", (unsigned long)dirent->stamp);
	}

	if (dirent->type == 'f') {
		fprintf(fh, "r,s%llu,\t%s\r\n", dirent->size, dirent->name);
	}
	else {
		fprintf(fh, "/,\t%s\r\n", dirent->name);
	}
}

void
statfilecmd (const char *filename)
{
	char *fullname = tf_makename(curdir(), filename);
	tf_dirent dirent;

	lreply (211, "status of %s:", filename);

	if (tf_stat(&tf, fullname, &dirent) != 0) {
		printf("No such file: %s\r\n", filename);
	}
	else {
		dump_dirent(stdout, filename, &dirent);
		printf("\r\n");
	}
	reply (211, "End of Status");
}

void
statcmd (void)
{
	struct sockaddr_in *sin;
	u_char *a, *p;
	tf_size_result result;

	lreply (211, "%s FTP server status:", hostname);
	if (!no_version)
		printf ("     ftpd (%s) %s\r\n",
			PACKAGE_NAME, PACKAGE_VERSION);
	printf ("     Connected to %s", cred.remotehost);
	if (!isdigit (cred.remotehost[0]))
		printf (" (%s)", inet_ntoa (his_addr.sin_addr));
	printf ("\r\n");
	if (cred.logged_in)
		{
			if (cred.guest)
	printf ("     Logged in anonymously\r\n");
			else
	printf ("     Logged in as %s\r\n", cred.name);
		}
	else if (askpasswd)
		printf ("     Waiting for password\r\n");
	else
		printf ("     Waiting for user name\r\n");

	if (tf_cmd_size(&tf, &result) == 0) {
		printf("     HDD Free  %10u kiB %7u MiB %4u GiB\r\n", result.freek, result.freek / 1024,
			   result.freek / (1024 * 1024));
		printf("     HDD Total %10u kiB %7u MiB %4u GiB\r\n", result.totalk, result.totalk / 1024,
			   result.totalk / (1024 * 1024));
	}
	else {
		printf ("     Unable to get HDD size info\r\n");
	}
	printf ("     Turbo mode is %s\r\n", turbo ? "ENABLED" : "DISABLED");
	printf ("     TYPE: %s", typenames[type]);
	if (type == TYPE_A || type == TYPE_E)
		printf (", FORM: %s", formnames[form]);
	if (type == TYPE_L)
#ifdef CHAR_BIT
		printf (" %d", CHAR_BIT);
#else
#if NBBY == 8
	printf (" %d", NBBY);
#else
	printf (" %d", bytesize);	/* need definition! */
#endif
#endif
	printf ("; STRUcture: %s; transfer MODE: %s\r\n",
		strunames[stru], modenames[stru_mode]);
	if (data != -1)
		printf ("     Data connection open\r\n");
	else if (pdata != -1)
		{
			printf ("     in Passive mode");
			sin = &pasv_addr;
			goto printaddr;
		}
	else if (usedefault == 0)
		{
			printf ("     PORT");
			sin = &data_dest;
		printaddr:
			a = (u_char *) &sin->sin_addr;
			p = (u_char *) &sin->sin_port;
#define UC(b) (((int) b) & 0xff)
			printf (" (%d,%d,%d,%d,%d,%d)\r\n", UC(a[0]),
				UC(a[1]), UC(a[2]), UC(a[3]), UC(p[0]), UC(p[1]));
#undef UC
		}
	else
		printf ("     No data connection\r\n");
	reply (211, "End of status");
}

void
fatal (const char *s)
{
	reply (451, "Error in server: %s\n", s);
	reply (221, "Closing connection due to server error.");
	dologout (0);
	/* NOTREACHED */
}

void
reply (int n, const char *fmt, ...)
{
	va_list ap;
#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
	va_start (ap, fmt);
#else
	va_start (ap);
#endif
	(void)printf ("%d ", n);
	(void)vprintf (fmt, ap);
	(void)printf ("\r\n");
	(void)fflush (stdout);
	if (debug)
		{
			syslog (LOG_DEBUG, "<--- %d ", n);
#ifdef HAVE_VSYSLOG
			vsyslog (LOG_DEBUG, fmt, ap);
#endif
		}
}

void
lreply (int n, const char *fmt, ...)
{
	va_list ap;
#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
	va_start (ap, fmt);
#else
	va_start (ap);
#endif
	(void)printf ("%d- ", n);
	(void)vprintf (fmt, ap);
	(void)printf ("\r\n");
	(void)fflush (stdout);
	if (debug)
		{
			syslog (LOG_DEBUG, "<--- %d- ", n);
#ifdef HAVE_VSYSLOG
			vsyslog (LOG_DEBUG, fmt, ap);
#endif
		}
}

static void
ack (const char *s)
{
	reply (250, "%s command successful.", s);
}

void
nack (const char *s)
{
	reply (502, "%s command not implemented.", s);
}

void
delete (const char *name)
{
	char *fullname = tf_makename(curdir(), name);

	LOGCMD ("delete", name);

	tf_cmd_delete(&tf, fullname);

	if (tf.error) {
		reply(550, "%s: %s.", name, tf_strerror(tf.error));
	}
	else {
		ack ("DELE");
	}
}

void
makedir (const char *name)
{
	char *fullname = tf_makename(curdir(), name);

	LOGCMD ("mkdir", name);

	tf_cmd_mkdir(&tf, fullname);

	if (tf.error) {
		reply(550, "%s: %s.", name, tf_strerror(tf.error));
		perror_reply (550, name);
	}
	else {
		reply (257, "\"%s\" new directory created.", fullname);
	}
}

void
removedir (const char *name)
{
	char *fullname = tf_makename(curdir(), name);

	LOGCMD ("rmdir", name);

	tf_cmd_delete(&tf, fullname);

	if (tf.error) {
		reply(550, "%s: %s.", name, tf_strerror(tf.error));
	}
	else {
		ack("RMD");
	}
}

void
pwd (void)
{
		reply (257, "\"%s\" is current directory.", curdir());
}

char *
renamefrom (const char *name)
{
	tf_dirent dirent;

	char *fullname = tf_makename(curdir(), name);

	if (tf_stat(&tf, fullname, &dirent) != 0) {
			reply (550, name);
			return 0;
	}
	reply (350, "File exists, ready for destination name");
	return (char *)(name);
}

void
renamecmd (const char *from, const char *to)
{
	LOGCMD2 ("rename", from, to);

	char *frompath = strdup(tf_makename(curdir(), from));
	char *topath = strdup(tf_makename(curdir(), to));

	/*fprintf(stderr, "tf_cmd_rename: frompath=%s, topath=%s\n", frompath, topath);*/
	if (tf_cmd_rename (&tf, frompath, topath) != 0)
		perror_reply (550, "rename");
	else {
		ack ("RNTO");
	}

	free(topath);
	free(frompath);
}

static void
dolog (struct sockaddr_in *sin, struct credentials *pcred)
{
	const char *name;
	struct hostent *hp = gethostbyaddr ((char *)&sin->sin_addr,
							sizeof (struct in_addr), AF_INET);

	if (hp)
		name = hp->h_name;
	else
		name = inet_ntoa (sin->sin_addr);

	if (pcred->remotehost)
		free (pcred->remotehost);
	pcred->remotehost = sgetsave (name);

#ifdef HAVE_SETPROCTITLE
	snprintf (proctitle, sizeof (proctitle), "%s: connected", pcred->remotehost);
	setproctitle ("%s",proctitle);
#endif /* HAVE_SETPROCTITLE */

	if (logging)
		syslog (LOG_INFO, "connection from %s", pcred->remotehost);
}

/*  Record logout in wtmp file
		and exit with supplied status.  */
void
dologout (int status)
{
	/* Racing condition with SIGURG: If SIGURG is receive
		 here, it will jump back has root in the main loop
		 David Greenman:dg@root.com.  */
	transflag = 0;

	topfield_close(&tf);

	/* beware of flushing buffers after a SIGPIPE */
	_exit (status);
}

static void
myoob (int signo)
{
	char *cp;

	(void)signo;
	/* only process if transfer occurring */
	if (!transflag)
		return;
	cp = tmpline;
	if (telnet_fgets (cp, 7, stdin) == NULL)
		{
			reply (221, "You could at least say goodbye.");
			dologout (0);
		}
	upper (cp);
	if (strcmp (cp, "ABOR\r\n") == 0)
		{
			tmpline[0] = '\0';
			reply (426, "Transfer aborted. Data connection closed.");
			reply (226, "Abort successful");
			longjmp (urgcatch, 1);
		}
	if (strcmp (cp, "STAT\r\n") == 0)
		{
			if (file_size != (off_t) -1)
	reply (213, "Status: %Ld of %Ld bytes transferred",
				 byte_count, file_size);
			else
	reply (213, "Status: %Ld bytes transferred",
				 byte_count);
		}
}

/* Note: a response of 425 is not mentioned as a possible response to
	 the PASV command in RFC959. However, it has been blessed as
	 a legitimate response by Jon Postel in a telephone conversation
	 with Rick Adams on 25 Jan 89.  */
void
passive (void)
{
	int len;
	char *p, *a;

	pdata = socket (AF_INET, SOCK_STREAM, 0);
	if (pdata < 0)
		{
			perror_reply (425, "Can't open passive connection");
			return;
		}
	pasv_addr = ctrl_addr;
	pasv_addr.sin_port = 0;
	(void) seteuid ((uid_t)0);
	if (bind (pdata, (struct sockaddr *)&pasv_addr, sizeof (pasv_addr)) < 0)
		{
			(void) seteuid ((uid_t)cred.uid);
			goto pasv_error;
		}
	(void) seteuid ((uid_t)cred.uid);
	len = sizeof(pasv_addr);
	if (getsockname (pdata, (struct sockaddr *) &pasv_addr, &len) < 0)
		goto pasv_error;
	if (listen (pdata, 1) < 0)
		goto pasv_error;
	a = (char *) &pasv_addr.sin_addr;
	p = (char *) &pasv_addr.sin_port;

#define UC(b) (((int) b) & 0xff)

	reply (227, "Entering Passive Mode (%d,%d,%d,%d,%d,%d)", UC(a[0]),
	 UC(a[1]), UC(a[2]), UC(a[3]), UC(p[0]), UC(p[1]));
	return;

 pasv_error:
	(void) close (pdata);
	pdata = -1;
	perror_reply (425, "Can't open passive connection");
	return;
}

/*
 * Format and send reply containing system error number.
 */
void
perror_reply (int code, const char *string)
{
	if (tf.error) {
		reply (code, "%s: %s.", string, tf_strerror (tf.error));
	}
	else {
		reply (code, "%s", string);
	}
}

void
send_file_list (const char *whichf, int LIST)
{
	tf_dir_entries entries;
	tf_dirent dirent;
	const char *pattern = 0;
	int exact = 0;
	FILE *dout = 0;
	int count = 0;
	int sort_type = TF_SORT_NAME;
	int ret;
	char *fullname;

	/* Now whichf may be a directory name, a file name, a wildcard pattern or NULL or empty
	 * And it may be one of the above with an initial sort specification
	 * such as -S, -tr, etc.
	 */

	pattern = whichf;

#ifdef DEBUG
	syslog(LOG_INFO, "send_file_list(whichf=%s), LIST=%d", whichf, LIST);
#endif

	if (pattern && *pattern == '-') {
		int reverse = 0;

		pattern++;
		while (*pattern && *pattern != ' ') {
			switch (*pattern) {
				case 't':
				case 'u':
				case 'c':
					sort_type = TF_SORT_TIME;
					break;

				case 'S':
					sort_type = TF_SORT_SIZE;
					break;

				case 'r':
					reverse = 1;
					break;
			}
			pattern++;
		}

		if (reverse) {
			sort_type = -sort_type;
		}

		while (*pattern == ' ') {
			pattern++;
		}
	}

#ifdef DEBUG
	syslog(LOG_INFO, "send_file_list: pattern=%s, sort_type=%d", pattern, sort_type);
#endif
	
	/* See if the specified filename exists */
	fullname = tf_makename(curdir(), pattern);

	if (pattern && *pattern) {
		/* If a parameter was given, what was it? */
		if (tf_stat(&tf, fullname, &dirent) == 0) {
			/* Exact file match, so list this one file or directory */
			exact = 1;
			fullname = tf_makename(curdir(), 0);
		}
		else {
			/* Does not exist, so treat whichf as a pattern */
			fullname = tf_makename(curdir(), 0);
		}
	}
	else {
#ifdef DEBUG
		syslog(LOG_INFO, "send_file_list: no pattern, so looking in dir %s", fullname);
#endif
	}

	ret = tf_cmd_dir_first(&tf, fullname, &entries);

	if (ret < 0) {
		reply(550, "%s: %s.", whichf, tf_strerror(ret));
		return;
	}

	if (entries.count > 0) {
		dout = dataconn("file list", (off_t)-1, "w");
		if (dout != NULL) {
			while (ret == 0) {
				int i;

				if (LIST) {
					tf_sort_dirents(&entries, sort_type);
				}

#ifdef DEBUG
				syslog(LOG_INFO, "send_file_list: examining %d entries with exact=%d and pattern=%s", entries.count, exact, pattern);
#endif

				for (i = 0; i < entries.count; i++) {
					if (pattern && *pattern) {
						if (exact && strcmp(pattern, entries.entry[i].name) != 0) {
							continue;
						}
						if (!exact && fnmatch(pattern, entries.entry[i].name, 0) != 0) {
							continue;
						}
					}
					count++;
					if (LIST) {
						if (use_eplf) {
							dump_dirent_as_eplf(dout, 0, &entries.entry[i]);
						}
						else {
							dump_dirent(dout, 0, &entries.entry[i]);
							fprintf(dout, "\r\n");
						}
					}
					else {
						fprintf(dout, "%s\r\n", entries.entry[i].name);
					}
				}
				ret = tf_cmd_dir_next(&tf, &entries);
			}
		}
	}

	if (count == 0) {
		reply (550, "No files found.");
	}
	else if (ferror (dout) != 0) {
		perror_reply (550, "Data connection");
	}
	else {
		reply (226, "Transfer complete.");
	}

	if (dout != NULL) {
		fclose (dout);
	}
	data = -1;
	pdata = -1;
}
